import unittest
import numpy
import gc
from dcmt import create_mts, create_mts_stripped, init_mt, rand, DcmtParameterError

class TestErrors(unittest.TestCase):

	def testWordlen(self):

		# correct
		for wordlen in (31, 32):
			create_mts(wordlen=wordlen, seed=1)

		# incorrect
		for wordlen in (16, 64):
			self.assertRaises(DcmtParameterError, create_mts, wordlen=wordlen, seed=1)

	def testExponent(self):

		#correct_exponents = [521, 607, 1279, 2203, 2281, 3217, 4253, 4423,
		#	9689, 9941, 11213, 19937, 21701, 23209, 44497]

		# small list for testing purposes, since the whole list takes
		# ages to process
		correct_exponents_reduced = [521, 607]

		# correct
		for exponent in correct_exponents_reduced:
			create_mts(exponent=exponent, seed=1)

		# incorrect
		for exponent in (500, 1000, 50000):
			self.assertRaises(DcmtParameterError, create_mts, exponent=exponent, seed=1)

	def testId(self):

		# error: start_id < max_id
		self.assertRaises(DcmtParameterError, create_mts, start_id=1, max_id=0, seed=1)

		# error: start_id < 0
		self.assertRaises(DcmtParameterError, create_mts, start_id=-1, max_id=2, seed=1)

		# error: start_id > 65536
		self.assertRaises(DcmtParameterError, create_mts, start_id=65537, max_id=65538, seed=1)

		# error: max_id > 65536
		self.assertRaises(DcmtParameterError, create_mts, start_id=65534, max_id=65538, seed=1)

	def testBugId9(self):

		# known bug: cannot find generator for wordlen=31, exponent=521 and id=9
		self.assertRaises(DcmtParameterError, create_mts, wordlen=31, exponent=521,
			start_id=7, max_id=10, seed=1)

	def testSeed(self):

		# since keywords are processed in extension, test that even not mentioning
		# keyword parameter works
		create_mts()

		# correct seeds
		for seed in (None, 0, 1, 2**16, long(2**16), 2**32 - 1, long(2**32 - 1)):
			mts = create_mts(seed=seed)
			init_mt(mts[0], seed=seed)

		# incorrect seeds
		for seed in (-1, -2**32 / 2, -2**32, 2**32, 2 * 2**32, long(2**32)):
			self.assertRaises(DcmtParameterError, create_mts, seed=seed)

			mts = create_mts()
			self.assertRaises(DcmtParameterError, init_mt, mts[0], seed=seed)


class TestBasics(unittest.TestCase):

	def testDifferentIds(self):
		mts = create_mts(start_id=0, max_id=1, seed=10)
		init_mt(mts[0], seed=2)
		init_mt(mts[1], seed=3)

		shape = (10,)
		randoms0 = rand(mts[0], shape)
		randoms1 = rand(mts[1], shape)

		for r0, r1 in zip(randoms0, randoms1):
			self.assertNotEqual(r0, r1)

	def testDeterminism(self):
		gen_id = 10
		gen_seed = 100
		init_seed = 300

		random_sets = []
		for i in xrange(2):
			mts = create_mts(start_id=gen_id, max_id=gen_id, seed=gen_seed)
			init_mt(mts[0], seed=init_seed)
			randoms = rand(mts[0], (10,))
			random_sets.append(randoms)

		for r0, r1 in zip(random_sets[0], random_sets[1]):
			self.assertEqual(r0, r1)

	def testRange(self):

		for wordlen in (31, 32):
			mts = create_mts(wordlen=wordlen, seed=99)
			init_mt(mts[0])

			r = rand(mts[0], (64,))
			self.assert_((r >= 0).all() and (r <= 2**wordlen - 1).all())


class TestStripped(unittest.TestCase):

	def testValidity(self):
		"""
		Check that common parameters are really common,
		and unique parameters are filled properly
		"""

		params = dict(start_id=0, max_id=2, seed=123)
		mts1 = create_mts(**params)
		mts2_common, mts2_stripped = create_mts_stripped(**params)

		for mt1, mt2 in zip(mts1, mts2_stripped):

			# check common parameters
			for field in (x[0] for x in mts2_common._fields_):
				self.assertEqual(getattr(mt1, field), getattr(mts2_common, field))

			# check unique parameters
			for field in (x[0] for x in mt2._fields_):
				self.assertEqual(getattr(mt1, field), getattr(mt2, field))


if __name__ == '__main__':

	suites = []

	for cls in (TestErrors, TestBasics, TestStripped):
		suites.append(unittest.TestLoader().loadTestsFromTestCase(cls))

	all_tests = unittest.TestSuite(suites)
	unittest.TextTestRunner(verbosity=1).run(all_tests)

	gc.collect() # additional test for different pointer- and reference- related bugs



import numpy
from time import time
from ctypes import Structure, POINTER, c_int, c_uint32, addressof, sizeof

from _libdcmt import create_mt_structs, fill_mt_structs, \
	fill_mt_structs_stripped, init_mt_struct, fill_rand_int, free_mt_structs
from .exceptions import DcmtParameterError


# matches structure definitions in C extension
class mt_common(Structure):
	_fields_ = [
		("mm", c_int),
		("nn", c_int),
		("rr", c_int),
		("ww", c_int),
		("wmask", c_uint32),
		("umask", c_uint32),
		("lmask", c_uint32),
		("shift0", c_int),
		("shift1", c_int),
		("shiftB", c_int),
		("shiftC", c_int)
	]

# matches structure definitions in C extension
class mt_unique(Structure):
	_fields_ = [
		("aaa", c_uint32),
		("maskB", c_uint32),
		("maskC", c_uint32)
	]


def validate_seed(seed):
	"""Return valid seed or raise an exception"""

	if seed is None:
		return int(time()) % (2 ** 32)
	elif seed < 0 or seed > 2 ** 32 - 1:
		raise DcmtParameterError("Seed must be between 0 and 2^32-1")

	return seed

def validate_parameters(wordlen=32, exponent=521, start_id=0, max_id=0, seed=None):
	"""Return valid parameter or raise an exception"""

	if start_id > max_id:
		raise DcmtParameterError("Starting ID must be equal to or lower than maximum ID")

	if start_id < 0 or max_id > 65535:
		raise DcmtParameterError("All IDs must lie between 0 and 65535")

	if wordlen == 31 and exponent == 521 and start_id <= 9 and max_id >= 9:
		raise DcmtParameterError("Known bug: cannot create generator for wordlen=31, exponent=521, id=9")

	if wordlen not in (31, 32):
		raise DcmtParameterError("Word length must be equal to 31 or 32")

	if exponent not in (521, 607, 1279, 2203, 2281, 3217, 4253, 4423,
			9689, 9941, 11213, 19937, 21701, 23209, 44497):
		raise DcmtParameterError("Mersenne exponent is incorrect or not from a supported set")

	return wordlen, exponent, start_id, max_id, validate_seed(seed)

def create_mts(**kwds):
	"""Create array of ctypes Structures with RNG parameters"""

	wordlen, exponent, start_id, max_id, seed = validate_parameters(**kwds)

	# This formula is not exported in dcmt API,
	# so I had to take it from seive.c::init_mt_search()
	state_len = exponent / wordlen + 1

	# Original library allocates two separate pieces of memory on a heap,
	# because the size of mt_struct depends on state vector size.
	# Since I will not use dcmt functions to delete mt_struct instances,
	# it is more convenient to keep everything in one struct,
	# using dynamic features of Python to build necessary class.
	class mt_struct(Structure):
		_fields_ = [
			("aaa", c_uint32),
			("mm", c_int),
			("nn", c_int),
			("rr", c_int),
			("ww", c_int),
			("wmask", c_uint32),
			("umask", c_uint32),
			("lmask", c_uint32),
			("shift0", c_int),
			("shift1", c_int),
			("shiftB", c_int),
			("shiftC", c_int),
			("maskB", c_uint32),
			("maskC", c_uint32),

			# filled on generator initialization
			("i", c_int),
			("state", POINTER(c_uint32)),

			# not present in original mt_struct
			("state_vec", c_uint32 * state_len)
		]

	ptr, count = create_mt_structs(wordlen, exponent, start_id, max_id, seed)

	# being extra-paranoidal
	try:
		mts = (mt_struct * count)()
		fill_mt_structs(addressof(mts), sizeof(mt_struct), mt_struct.state_vec.offset, ptr, count)
	finally:
		free_mt_structs(ptr, count)

	return mts

def init_mt(mt, seed=None):
	"""Initialize RNG structure with given seed"""
	seed = validate_seed(seed)
	init_mt_struct(addressof(mt), seed)

def rand(mt, shape):
	"""Return numpy array with random numbers"""
	randoms = numpy.empty(shape, dtype=numpy.uint32)
	fill_rand_int(addressof(mt), randoms)
	return randoms

def create_mts_stripped(**kwds):
	"""Return optimized RNG structures"""
	wordlen, exponent, start_id, max_id, seed = validate_parameters(**kwds)
	ptr, count = create_mt_structs(wordlen, exponent, start_id, max_id, seed)

	# being extra-paranoidal
	try:
		mts_common = mt_common()
		mts_unique = (mt_unique * count)()
		fill_mt_structs_stripped(addressof(mts_common), addressof(mts_unique), ptr, count)
	finally:
		free_mt_structs(ptr, count)

	return mts_common, mts_unique
